#include"t10n_lexer.hpp"
#include"t10n_block.hpp"




namespace t10n{




void
Lexer::
newline()
{
  position_.column_number = 0;

  position_.line_number += 1;
  pointer_              += 1;
}


void
Lexer::
advance(uint32_t  n)
{
  position_.column_number += n;
  pointer_                += n;
}


int
to_char(int  c)
{
    switch(c)
    {
      case('t' ): return'\t';
      case('n' ): return'\n';
      case('0' ): return'\0';
      case('r' ): return'\r';
      case('\\'): return'\\';
      case('\''): return'\'';
      case('\"'): return'\"';
    }


  return -1;
}


void
Lexer::
read_str_literal(const Position&  pos, TokenType  type, const int  close_chr)
{
  buffer_.clear();

    for(;;)
    {
      int  c = *pointer_;

        if(!c)
        {
          report;
          return;
        }


        if(c == close_chr)
        {
          advance();

          tokens_.emplace_back(pos,type,buffer_);

          break;
        }

      else
        if(c == '\\')
        {
          advance();

          c = to_char(*pointer_);

            if(c == -1)
            {
              report;
              return;
            }
        }


      buffer_.push_back(c);

      advance();
    }
}


void
Lexer::
read_identifier(const Position&  pos, int  first_c)
{
  buffer_.clear();

  buffer_.append(1,first_c);

    for(;;)
    {
      const int  c = *pointer_;

        if(isalnum(c) ||
           (c == '_'))
        {
          advance();

          buffer_.push_back(c);
        }

      else
        {
          break;
        }
    }


  tokens_.emplace_back(pos,TokenType::identifier,buffer_);
}


bool
isop(int  c)
{
  return((c == '+') ||
         (c == '-') ||
         (c == '*') ||
         (c == '/') ||
         (c == '%') ||
         (c == '&') ||
         (c == '|') ||
         (c == '^') ||
         (c == '~') ||
         (c == '!') ||
         (c == '=') ||
         (c == '>') ||
         (c == '<') ||
         (c == '.') ||
         (c == ',') ||
         (c == ';') ||
         (c == ':') ||
         (c == '?') ||
         (c == '#'));
}


void
Lexer::
read_operator(const Position&  pos)
{
  Operator  op;

       if(strncmp(pointer_,"<<=",3) == 0){op = Operator('<','<','=');}
  else if(strncmp(pointer_,">>=",3) == 0){op = Operator('>','>','=');}
  else if(strncmp(pointer_,"...",3) == 0){op = Operator('.','.','.');}
  else if(strncmp(pointer_,"+=",2) == 0){op = Operator('+','=');}
  else if(strncmp(pointer_,"-=",2) == 0){op = Operator('-','=');}
  else if(strncmp(pointer_,"*=",2) == 0){op = Operator('*','=');}
  else if(strncmp(pointer_,"/=",2) == 0){op = Operator('/','=');}
  else if(strncmp(pointer_,"%=",2) == 0){op = Operator('%','=');}
  else if(strncmp(pointer_,"^=",2) == 0){op = Operator('^','=');}
  else if(strncmp(pointer_,"|=",2) == 0){op = Operator('|','=');}
  else if(strncmp(pointer_,"&=",2) == 0){op = Operator('&','=');}
  else if(strncmp(pointer_,"==",2) == 0){op = Operator('=','=');}
  else if(strncmp(pointer_,"!=",2) == 0){op = Operator('!','=');}
  else if(strncmp(pointer_,"<=",2) == 0){op = Operator('<','=');}
  else if(strncmp(pointer_,">=",2) == 0){op = Operator('>','=');}
  else if(strncmp(pointer_,"&&",2) == 0){op = Operator('&','&');}
  else if(strncmp(pointer_,"||",2) == 0){op = Operator('|','|');}
  else if(strncmp(pointer_,"->",2) == 0){op = Operator('-','>');}
  else if(strncmp(pointer_,"++",2) == 0){op = Operator('+','+');}
  else if(strncmp(pointer_,"--",2) == 0){op = Operator('-','-');}
  else if(strncmp(pointer_,"::",2) == 0){op = Operator(':',':');}
  else if(strncmp(pointer_,"<<",2) == 0){op = Operator('<','<');}
  else if(strncmp(pointer_,">>",2) == 0){op = Operator('>','>');}
  else if(strncmp(pointer_,"##",2) == 0){op = Operator('#','#');}
  else if(strncmp(pointer_,"+",1) == 0){op = Operator('+');}
  else if(strncmp(pointer_,"-",1) == 0){op = Operator('-');}
  else if(strncmp(pointer_,"*",1) == 0){op = Operator('*');}
  else if(strncmp(pointer_,"/",1) == 0){op = Operator('/');}
  else if(strncmp(pointer_,"%",1) == 0){op = Operator('%');}
  else if(strncmp(pointer_,"&",1) == 0){op = Operator('&');}
  else if(strncmp(pointer_,"|",1) == 0){op = Operator('|');}
  else if(strncmp(pointer_,"^",1) == 0){op = Operator('^');}
  else if(strncmp(pointer_,"~",1) == 0){op = Operator('~');}
  else if(strncmp(pointer_,"!",1) == 0){op = Operator('!');}
  else if(strncmp(pointer_,"=",1) == 0){op = Operator('=');}
  else if(strncmp(pointer_,"<",1) == 0){op = Operator('<');}
  else if(strncmp(pointer_,">",1) == 0){op = Operator('>');}
  else if(strncmp(pointer_,".",1) == 0){op = Operator('.');}
  else if(strncmp(pointer_,",",1) == 0){op = Operator(',');}
  else if(strncmp(pointer_,":",1) == 0){op = Operator(':');}
  else if(strncmp(pointer_,";",1) == 0){op = Operator(';');}
  else if(strncmp(pointer_,"#",1) == 0){op = Operator('#');}
  else if(strncmp(pointer_,"?",1) == 0){op = Operator('?');}
  else
    {
      report;
      return;
    }


  tokens_.emplace_back(pos,op);

  advance(std::strlen(op.s));
}


Lexer::
Lexer(TokenList&  tokens, Position&  position, const char*&  pointer, const int  close):
pointer_(pointer),
position_(position),
tokens_(tokens)
{
    for(;;)
    {
      const int  c = *pointer_;

      Position  pos = position_;

        if(!c)
        {
          break;
        }

      else
        if(isspace(c))
        {
          skip_spaces();
        }

      else
        if(c == close)
        {
          advance();

          break;
        }

      else
        if(c == '(')
        {
          advance();

          auto  blk = new Block(Bracket('(',')'),position_,pointer_);

          tokens_.emplace_back(pos,blk);
        }

      else
        if(c == '[')
        {
          advance();

          auto  blk = new Block(Bracket('[',']'),position_,pointer_);

          tokens_.emplace_back(pos,blk);
        }

      else
        if(c == '{')
        {
          advance();

          auto  blk = new Block(Bracket('{','}'),position_,pointer_);

          tokens_.emplace_back(pos,blk);
        }

      else
        if(std::strncmp(pointer_,"//",2) == 0)
        {
          advance(2);

          skip_line_comment();
        }

      else
        if(std::strncmp(pointer_,"/*",2) == 0)
        {
          advance(2);

          skip_block_comment();
        }

      else
        if(c == '\'')
        {
          advance();

          read_str_literal(pos,TokenType::chr_literal,'\'');
        }

      else
        if(c == '\"')
        {
          advance();

          read_str_literal(pos,TokenType::str_literal,'\"');
        }

      else
        if(std::strncmp(pointer_,"u\'",2) == 0)
        {
          advance(2);

          read_str_literal(pos,TokenType::u16chr_literal,'\'');
        }

      else
        if(std::strncmp(pointer_,"u\"",2) == 0)
        {
          advance(2);

          read_str_literal(pos,TokenType::u16str_literal,'\"');
        }

      else
        if(isdigit(c))
        {
          read_int_literal(pos);
        }

      else
        if(isop(c))
        {
          read_operator(pos);
        }

      else
        if(isalpha(c) || (c == '_'))
        {
          advance();

          read_identifier(pos,c);
        }

      else
        {
          advance();

          printf("cannot process \'%c\'\n",c);
        }
    }
}




}




