#include"t10n_token.hpp"
#include"t10n_block.hpp"




namespace t10n{




Token::
Token(Position  pos, TokenType  t, const std::string&  buf):
position(pos),
type(t),
s(new std::string(buf))
{
}


Token::
Token(Position  pos, Operator  op):
position(pos),
type(TokenType::operator_),
i(op)
{
}


Token::
Token(Position  pos, TokenType  t, unsigned int  u):
position(pos),
type(t),
i(u)
{
}


Token::
Token(Position  pos, Block*  blk_):
position(pos),
type(TokenType::block),
blk(blk_)
{
}


Token::
Token(const Token&  rhs):
position(rhs.position),
type(rhs.type)
{
    switch(type)
    {
      case(TokenType::block):
        blk = new Block(*rhs.blk);
        break;
      case(TokenType::identifier):
      case(TokenType::chr_literal):
      case(TokenType::str_literal):
      case(TokenType::u16chr_literal):
      case(TokenType::u16str_literal):
      {
        s = new std::string(*rhs.s);
      } break;
      default:
        i = rhs.i;
    }
}


Token::
Token(Token&&  rhs):
position(rhs.position),
type(rhs.type)
{
  s = rhs.s          ;
      rhs.s = nullptr;
}


Token::
~Token()
{
  clear();
}




void
Token::
clear()
{
    switch(type)
    {
      case(TokenType::block):
        delete blk;
        break;
      case(TokenType::identifier):
      case(TokenType::chr_literal):
      case(TokenType::str_literal):
      case(TokenType::u16chr_literal):
      case(TokenType::u16str_literal):
        delete s;
        break;
    }


  type = TokenType::null;
}


void
Token::
change_to_keyword(int  n)
{
  clear();

  i = n;

  type = TokenType::keyword;
}


void
Token::
print(int  indent_n) const
{
  print_indent(indent_n);

  printf("[line:%4d, col:%4d] (",1+position.line_number,
                                 1+position.column_number);

    switch(type)
    {
      case(TokenType::chr_literal):
        printf("chr");
        break;
      case(TokenType::str_literal):
        printf("str");
        break;
      case(TokenType::u16chr_literal):
        printf("uchr");
        break;
      case(TokenType::u16str_literal):
        printf("ustr");
        break;
      case(TokenType::identifier):
        printf("identifier");
        break;
      case(TokenType::int_literal):
        printf("int");
        break;
      case(TokenType::operator_):
        printf("operator");
        break;
      case(TokenType::block):
        printf("block");
        break;
    }


  printf(")");

    switch(type)
    {
      case(TokenType::chr_literal):
        printf("\'");
        break;
      case(TokenType::str_literal):
        printf("\"");
        break;
      case(TokenType::u16chr_literal):
        printf("u\'");
        break;
      case(TokenType::u16str_literal):
        printf("u\"");
        break;
      case(TokenType::identifier):
        printf("%s",s->data());
        break;
      case(TokenType::int_literal):
        printf("%d",i);
        break;
      case(TokenType::operator_):
      {
        Operator  op;

        op = i;

        printf("%s",op.s);
      }

        break;
      case(TokenType::block):
        printf("\n");
        blk->print(indent_n+2);
        break;
    }


  const char*  p = s->data();

    switch(type)
    {
      case(TokenType::chr_literal):
      case(TokenType::str_literal):
      case(TokenType::u16chr_literal):
      case(TokenType::u16str_literal):
        while(*p)
        {
          int  c = *p++;

            switch(c)
            {
              case('\0'): printf("\\0");break;
              case('\t'): printf("\\t");break;
              case('\n'): printf("\\n");break;
              case('\\'): printf("\\\\");break;
              case('\''): printf("\\\'");break;
              case('\"'): printf("\\\"");break;
              default:
                fputc(c,stdout);
            }
        }
        break;
    }


    switch(type)
    {
      case(TokenType::chr_literal):
      case(TokenType::u16chr_literal):
        printf("\'");
        break;
      case(TokenType::str_literal):
      case(TokenType::u16str_literal):
        printf("\"");
        break;
    }
}




void
print_indent(int  n)
{
    while(n--)
    {
      printf(" ");
    }
}


}




