#ifndef T10N_TOKEN_HPP_INCLUDED
#define T10N_TOKEN_HPP_INCLUDED


#include<cstdint>
#include<cstddef>
#include<cstring>
#include<cctype>
#include<string>
#include<list>
#include"t10n_operator.hpp"


namespace t10n{


enum class
TokenType
{
  null,

     chr_literal,
  u16chr_literal,
     str_literal,
  u16str_literal,
     int_literal,

  operator_,

  keyword,

  identifier,

  block,

};


struct
Position
{
  uint32_t    line_number=0;
  uint32_t  column_number=0;

};


class Block;


struct
Token
{
  const Position  position;

  TokenType  type;

  union{
    unsigned int  i;
    std::string*  s;

    Block*  blk;
  };


  Token(Position  pos, TokenType  t, const std::string&  buf);
  Token(Position  pos, TokenType  t, unsigned int  u);
  Token(Position  pos, Block*  blk_);
  Token(Position  pos, Operator  op);
  Token(const Token&   rhs);
  Token(      Token&&  rhs);
 ~Token();


  void  clear();

  void  change_to_keyword(int  n);

  void  print(int  indent_n) const;

};


void  print_indent(int  n);


}




#endif




